import astropy.units as u

import roentgen

__all__ = [
    "is_an_element",
    "get_atomic_number",
    "is_in_known_compounds",
    "get_compound_index",
    "get_material_density",
    "get_element_symbol",
    "density_ideal_gas",
    "get_material_name",
    "get_material_symbol",
]


def is_an_element(element_str):
    """Returns True if the string represents an element"""
    result = False
    lower_case_list = list([s.lower() for s in roentgen.elements["symbol"]])
    if (len(element_str) <= 2) and (element_str.lower() in lower_case_list):
        result = True
    else:
        lower_case_list = list([s.lower() for s in roentgen.elements["name"]])
        if element_str.lower() in lower_case_list:
            result = True
    return result


def get_element_symbol(element_str):
    """Return the element abbreviation"""
    lower_case_symbol_list = list([s.lower() for s in roentgen.elements["symbol"]])
    if element_str.lower() in lower_case_symbol_list:  # already a symbol
        return element_str.capitalize()
    elif is_an_element(element_str):
        lower_case_list = list([s.lower() for s in roentgen.elements["name"]])
        return roentgen.elements[lower_case_list.index(element_str.lower())]["symbol"].capitalize()
    else:
        raise ValueError(f"{element_str} not recognized.")


def get_atomic_number(element_str):
    """Return the atomic number of the element"""
    # check to see if element_str is symbol
    if is_an_element(element_str):
        if len(element_str) <= 2:
            lower_case_list = list([s.lower() for s in roentgen.elements["symbol"]])
            atomic_number = roentgen.elements[lower_case_list.index(element_str.lower())]["z"]
        else:
            lower_case_list = list([s.lower() for s in roentgen.elements["name"]])
            atomic_number = roentgen.elements[lower_case_list.index(element_str.lower())]["z"]
    else:
        raise ValueError(f"{element_str} not recognized.")
    return atomic_number


def is_in_known_compounds(compound_str):
    """Returns True is the compound is in the list of known compounds"""
    lcase_symbols_list = list([s.lower() for s in roentgen.compounds["symbol"]])
    lcase_name_list = list([s.lower() for s in roentgen.compounds["name"]])
    case1 = compound_str.lower() in lcase_symbols_list
    case2 = compound_str.lower() in lcase_name_list
    return case1 or case2


def get_compound_index(compound_str):
    """Return the index of the compound in the compound table"""
    if is_in_known_compounds(compound_str):
        lower_case_symbols_list = list([s.lower() for s in roentgen.compounds["symbol"]])
        if compound_str.lower() in lower_case_symbols_list:
            return lower_case_symbols_list.index(compound_str.lower())
        lcase_name_list = list([s.lower() for s in roentgen.compounds["name"]])
        if compound_str.lower() in lcase_name_list:
            return lcase_name_list.index(compound_str.lower())
    else:
        raise ValueError(f"{compound_str} not recognized.")


def get_material_value(material_str, column_str):
    if is_an_element(material_str):
        ind = get_atomic_number(material_str) - 1
        return roentgen.elements[ind][column_str]
    elif is_in_known_compounds(material_str):
        # not using loc because table indexing is not yet stable
        # density = roentgen.compounds.loc[material_str]['density']
        index = get_compound_index(material_str)
        return roentgen.compounds[index][column_str]
    else:
        raise ValueError(f"{material_str} not recognized.")


def get_material_symbol(material_str):
    """Given a material name return the symbol"""
    return get_material_value(material_str, "symbol")


def get_material_name(material_str):
    """Given a material name return the official name"""
    return get_material_value(material_str, "name")


def get_material_density(material_str):
    """Given a material name return the default density"""
    return get_material_value(material_str, "density")


@u.quantity_input(pressure=u.pascal, temperature=u.deg_C, equivalencies=u.temperature())
def density_ideal_gas(pressure, temperature) -> u.kg / u.m**3:
    """Given the pressure and temperature of a dry gas, return the density using the ideal gas law"""
    R = 287.058 * u.J / u.kg / u.Kelvin
    result = pressure / (R * temperature.to(u.K, equivalencies=u.temperature()))
    return result.decompose()



import numpy as np
import pytest

import astropy.units as u

import roentgen
from roentgen.absorption import Material, Response

all_materials = list(roentgen.elements["symbol"]) + list(roentgen.compounds["symbol"])
energy_array = u.Quantity(np.arange(1, 100, 1), "keV")
detector = Material("Si", 500 * u.micron)
thin_material = Material("air", thickness=1e-30 * u.um)
thick_material = Material("gold", thickness=1 * u.Mm)


@pytest.mark.parametrize("wrong_type", [(None, [1, 2, 3], 1, "Si")])
def test_response_detector_is_none_error(wrong_type):
    # check that the response class raises an error is not Stack or Material
    with pytest.raises(TypeError):
        Response(thin_material, detector=wrong_type)


def test_thin_detector():
    # check if a super thin material means a very small response
    resp = Response(optical_path=thin_material, detector=thin_material)
    assert resp.response(u.Quantity([1], "keV")) < 0.01


def test_thick_optical_path():
    # check if a super thick optical path means a very small response
    resp = Response(optical_path=thick_material, detector=thin_material)
    assert resp.response(u.Quantity([1], "keV")) < 0.01


def test_optical_path_stack_input():
    thin_material_stack = Material("air", thickness=1e-30 * u.um)
    for i in range(10):
        thin_material_stack += thin_material
    assert isinstance(Response(thin_material_stack, detector=detector), Response)


def test_bad_optical_path():
    with pytest.raises(TypeError):
        Response(optical_path="Si", detector=thin_material)



def get_package_data():
    return {
        "roentgen": [
            "data/*.json",
            "data/*.csv",
            "data/elements/*.csv",
            "data/compounds_mixtures/*.csv",
        ]
    }



import numpy as np
import pytest

import astropy.units as u
from astropy.units import cds, imperial

import roentgen
import roentgen.util as util

# this enables the atm unit
cds.enable()

all_materials = list(roentgen.elements["symbol"]) + list(roentgen.compounds["symbol"])
not_real_materials = ["adamantium", "ice-nine", "kryptonite", "redstone", "unobtainium"]


@pytest.mark.parametrize("element", roentgen.elements["symbol"])
def test_is_an_element_symbol(element):
    assert util.is_an_element(element)


@pytest.mark.parametrize("element", roentgen.elements["name"])
def test_is_an_element_name(element):
    assert util.is_an_element(element)


@pytest.mark.parametrize("element", not_real_materials)
def test_is_not_an_element(element):
    assert util.is_an_element(element) is False


@pytest.mark.parametrize("name,symbol", zip(roentgen.elements["name"], roentgen.elements["symbol"]))
def test_get_element_symbol(name, symbol):
    assert util.get_element_symbol(name) == symbol


@pytest.mark.parametrize("symbol", roentgen.elements["symbol"])
def test_get_element_symbol_same_return(symbol):
    assert util.get_element_symbol(symbol) == symbol


@pytest.mark.parametrize("element", not_real_materials)
def test_get_element_symbol_error(element):
    with pytest.raises(ValueError):
        util.get_element_symbol(element)


@pytest.mark.parametrize("symbol,z", zip(roentgen.elements["symbol"], roentgen.elements["z"]))
def test_get_atomic_number_symbol(symbol, z):
    assert util.get_atomic_number(symbol) == z


@pytest.mark.parametrize("name,z", zip(roentgen.elements["name"], roentgen.elements["z"]))
def test_get_atomic_number_name(name, z):
    assert util.get_atomic_number(name) == z


@pytest.mark.parametrize("element", not_real_materials)
def test_get_atomic_number_error(element):
    with pytest.raises(ValueError):
        util.get_atomic_number(element)


@pytest.mark.parametrize("symbol", roentgen.compounds["symbol"])
def test_is_in_known_compounds_symbol(symbol):
    assert util.is_in_known_compounds(symbol)


@pytest.mark.parametrize("name", roentgen.compounds["name"])
def test_is_in_known_compounds_name(name):
    assert util.is_in_known_compounds(name)


@pytest.mark.parametrize("element", not_real_materials)
def test_is_in_known_compounds_not(element):
    assert util.is_in_known_compounds(element) is False


@pytest.mark.parametrize("symbol", roentgen.compounds["symbol"])
def test_get_compound_index_symbol(symbol):
    assert util.get_compound_index(symbol) == roentgen.compounds.loc_indices[symbol]


@pytest.mark.parametrize("name", roentgen.compounds["name"])
def test_get_compound_index_name(name):
    assert util.get_compound_index(name) == roentgen.compounds["name"].tolist().index(name)


@pytest.mark.parametrize("element", not_real_materials)
def test_get_compound_index_error(element):
    with pytest.raises(ValueError):
        util.get_compound_index(element)


@pytest.mark.parametrize("element", not_real_materials)
def test_get_symbol_error(element):
    with pytest.raises(ValueError):
        util.get_material_symbol(element)


@pytest.mark.parametrize("element", not_real_materials)
def test_get_material_name(element):
    with pytest.raises(ValueError):
        util.get_material_name(element)


@pytest.mark.parametrize("element", not_real_materials)
def test_get_density_error(element):
    with pytest.raises(ValueError):
        util.get_material_density(element)


@pytest.mark.parametrize(
    "symbol,density", zip(roentgen.elements["symbol"], roentgen.elements["density"])
)
def test_get_density_element(symbol, density):
    assert util.get_material_density(symbol) == density


@pytest.mark.parametrize(
    "compound,density", zip(roentgen.compounds["symbol"], roentgen.compounds["density"])
)
def test_get_density_compound(compound, density):
    assert util.get_material_density(compound) == density


@pytest.mark.parametrize(
    "material", list(roentgen.compounds["symbol"]) + list(roentgen.compounds["symbol"])
)
def test_get_density_quantity(material):
    assert isinstance(util.get_material_density(material), u.Quantity)


@pytest.mark.parametrize(
    "pressure,temperature,result",
    [
        (1 * cds.atm, 15 * u.Celsius, 1.225 * u.kg / u.m**3),
        (101325 * u.pascal, 15 * u.Celsius, 1.225 * u.kg / u.m**3),
        (1 * cds.atm, 288.15 * u.K, 1.225 * u.kg / u.m**3),
        (1 * cds.atm, 59 * imperial.deg_F, 1.225 * u.kg / u.m**3),
        (5 * cds.atm, 20 * u.Celsius, 6.020 * u.kg / u.m**3),
    ],
)
def test_density_ideal_gas(pressure, temperature, result):
    assert np.isclose(util.density_ideal_gas(pressure, temperature), result, rtol=1e-4)


@pytest.mark.parametrize(
    "pressure,temperature",
    [
        (5 * u.m, 15 * u.Celsius),
        (5 * u.s, 15 * u.Celsius),
        (1 * cds.atm, 5 * u.m),
        (1 * cds.atm, 5 * u.gauss),
    ],
)
def test_density_ideal_gas_bad_units(pressure, temperature):
    with pytest.raises(u.UnitsError):
        util.density_ideal_gas(pressure, temperature)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""
A Python package for the quantitative analysis of the interaction of energetic
photons with matter (x-rays and gamma-rays).
"""
import os

import astropy.units as u
from astropy.io import ascii
from astropy.table import QTable, Table

try:
    from .version import __version__
except ImportError:
    __version__ = "unknown"
__all__ = []
# roentgen specific configuration
# load some data files on import

_package_directory = os.path.dirname(os.path.abspath(__file__))
_data_directory = os.path.abspath(os.path.join(_package_directory, "data"))

elements_file = os.path.join(_data_directory, "elements.csv")
elements = QTable(ascii.read(elements_file, format="csv"))

elements["density"].unit = u.g / (u.cm**3)
elements["i"].unit = u.eV
elements["ionization energy"].unit = u.eV
elements["atomic mass"] = elements["z"] / elements["zovera"] * u.u
elements.add_index("z")

compounds_file = os.path.join(_data_directory, "compounds_mixtures.csv")
compounds = QTable(ascii.read(compounds_file, format="csv", fast_reader=False))
compounds["density"].unit = u.g / (u.cm**3)
compounds.add_index("symbol")

notation_translation = Table(
    ascii.read(
        os.path.join(_data_directory, "siegbahn_to_iupac.csv"),
        format="csv",
        fast_reader=False,
    )
)

emission_lines = QTable(
    ascii.read(
        os.path.join(_data_directory, "emission_lines.csv"),
        format="csv",
        fast_reader=False,
    )
)
# not sure why i need to fix this otherwise it is \ufenergy
emission_lines.rename_column(emission_lines.colnames[0], "energy")
emission_lines[emission_lines.colnames[0]].unit = u.eV
emission_lines.add_index(emission_lines.colnames[0])
emission_lines.add_index(emission_lines.colnames[1])
emission_lines.meta = {
    "source": "Center for X-ray Optics and Advanced Light Source, X-Ray Data Booklet Table 1-2",
    "publication date": "2009 October",
    "url": "https://xdb.lbl.gov/Section1/Table_1-3.pdf",
}



import numpy as np
import pytest

import astropy.units as u

import roentgen
from roentgen.absorption.material import Material, Stack

all_materials = list(roentgen.elements["symbol"]) + list(roentgen.compounds["symbol"])
energy_array = u.Quantity(np.arange(1, 100, 1), "keV")


@pytest.fixture(params=all_materials)
def material(request):
    return Material(request.param, 500 * u.micron)


def test_twomaterials_to_stack(material):
    # check that adding two materials returns a compound
    assert isinstance(material + Material("Si", 500 * u.micron), Stack)


def test_threematerials_to_stack(material):
    # check that adding three materials returns a compound
    assert isinstance(
        material + Material("Ge", 500 * u.micron) + Material("cdte", 100 * u.micron), Stack
    )


def test_add_material_to_stack():
    stack1 = Material("Ge", 500 * u.micron) + Material("Ge", 100 * u.micron)
    assert isinstance(stack1 + Material("Ge", 500 * u.micron), Stack)


def test_add_two_stacks():
    material_str = ["Ge", "cdte", "Si", "Al"]
    stack1 = Material(material_str[0], 500 * u.micron) + Material(material_str[1], 100 * u.micron)
    stack2 = Material(material_str[2], 500 * u.micron) + Material(material_str[2], 500 * u.micron)
    total_stack = stack1 + stack2
    all_materials = [this_mat.symbol for this_mat in total_stack.materials]
    assert isinstance(total_stack, Stack)
    # check that the stack has all the same number of materials
    assert len(material_str) == len(total_stack.materials)
    # check that the stack has all the same material names
    assert all_materials.sort() == material_str.sort()


def test_stack_calculations(material):
    comp = Material("Ge", 500 * u.micron) + Material("cdte", 100 * u.micron)
    # test that it returns the same number of elements as energy array
    assert len(comp.absorption(energy_array)) == len(energy_array)
    assert len(comp.transmission(energy_array)) == len(energy_array)


def test_bad_add_to_stack():
    with pytest.raises(TypeError):
        Material("Ge", 500 * u.micron) + Material("Si", 100 * u.micron) + ["foo", "bar"]


def test_mat_add_stack():
    stack1 = Material("Ge", 500 * u.micron) + Material("Si", 100 * u.micron)
    assert isinstance(Material("Ge", 500 * u.micron) + stack1, Stack)



import pytest
from numpy.testing import assert_allclose

import astropy.units as u

from roentgen.absorption.material import MassAttenuationCoefficient

not_real_materials = ["adamantium", "ice-nine", "kryptonite", "redstone", "unobtainium"]


def test_interpolate_matches_at_data():
    for this_element in ["Te", "Si", "Ge", "cdte"]:
        te = MassAttenuationCoefficient(this_element)
        assert_allclose(te.data, te.func(te.energy))

        # now change one point and make sure it no longer works
        te.energy[0] = 10 * u.keV
        with pytest.raises(AssertionError):
            assert_allclose(te.data, te.func(te.energy))


@pytest.mark.parametrize("element", not_real_materials)
def test_nonexistent_material(element):
    with pytest.raises(ValueError):
        MassAttenuationCoefficient(element)



import numpy as np
import pytest

import astropy.units as u

import roentgen
from roentgen.absorption.material import MassAttenuationCoefficient, Material
from roentgen.util import get_material_density, is_an_element

all_materials = list(roentgen.elements["symbol"]) + list(roentgen.compounds["symbol"])
energy_array = u.Quantity(np.arange(1, 100, 1), "keV")


@pytest.fixture(params=all_materials)
def mass_atten(request):
    return MassAttenuationCoefficient(request.param)


def test_is_an_element_symbol():
    """Test that function is okay with all known elements symbols"""
    for el in roentgen.elements["symbol"]:
        assert is_an_element(el)


def test_is_an_element_caseinsensitive_symbol():
    """Test that searching for an element symbol is not case sensitive"""
    for el in roentgen.elements["symbol"]:
        assert is_an_element(el.upper())
        assert is_an_element(el.lower())
        assert is_an_element(el.capitalize())


def test_is_an_element_name():
    """Test that function is okay with all known elements names"""
    for el in roentgen.elements["name"]:
        assert is_an_element(el)


def test_is_an_element_caseinsensitive_name():
    """Test that searching for an element name is not case sensitive"""
    for el in roentgen.elements["name"]:
        assert is_an_element(el.upper())
        assert is_an_element(el.lower())
        assert is_an_element(el.capitalize())


def test_mass_atten(mass_atten):
    # check that all materials can load a MassAttenuationCoefficient object
    assert isinstance(mass_atten, MassAttenuationCoefficient)


def test_returns_quantity(mass_atten):
    assert isinstance(mass_atten.func(1 * u.keV), u.Quantity)


def test_number_of_energies(mass_atten):
    energy = u.Quantity(np.arange(1, 1000), "keV")
    atten = mass_atten.func(energy)
    assert len(energy) == len(atten)


@pytest.fixture(params=all_materials)
def material(request):
    return Material(request.param, 500 * u.micron)


def test_material(material):
    assert isinstance(material, Material)


def test_material_give_density():
    assert isinstance(
        Material(all_materials[40], 500 * u.micron, density=1 * u.kg / u.m**3), Material
    )


@pytest.mark.parametrize(
    "material_wrong",
    [(1), (["Si", "He"]), (4.5)],
)
def test_material_bad_input(material_wrong):
    with pytest.raises(TypeError):
        Material(material_wrong, 500 * u.micron)


def test_bad_add_to_material():
    with pytest.raises(TypeError):
        Material("Ge", 500 * u.micron) + ["foo", "bar"]


@pytest.fixture(params=all_materials)
def thick_material(request):
    return Material(request.param, 500 * u.Mm)


def test_opaque(thick_material):
    # check that extremely large amounts of material mean no transmission
    assert thick_material.transmission(1 * u.keV) < 1e-6


@pytest.fixture(params=all_materials)
def thin_material(request):
    return Material(request.param, 1e-6 * u.pm)


def test_transparent(thin_material):
    # check that extremely large amounts of material mean no transmission
    assert thin_material.transmission(1 * u.keV) > 0.90


@pytest.mark.parametrize(
    "material,energy,thickness",
    [
        ("Si", 10 * u.keV, 1 * u.mm),
        ("Au", 1 * u.keV, 1 * u.micron),
        ("Air (dry)", 1 * u.keV, 1 * u.m),
        ("blood", 5 * u.keV, 1 * u.mm),
    ],
)
def test_linear_attenuation_coefficient(material, energy, thickness):
    mat = Material(material, thickness)
    assert np.isclose(
        mat.linear_attenuation_coefficient(energy),
        mat.mass_attenuation_coefficient(energy) * get_material_density(material),
        rtol=1e-4,
    )


@pytest.mark.parametrize(
    "material",
    [
        ("Si"),
        ("Au"),
        ("Air (dry)"),
        ("blood"),
        ({"Fe": 0.98, "C": 0.02}),  # steel
        ({"Cu": 0.88, "Sn": 0.12}),  # bronze
        ({"water": 0.97, "Na": 0.015, "Cl": 0.015}),  # salt water
    ],
)
def test_material_number_of_energies(material):
    mat = Material(material, 1 * u.m)
    energy = u.Quantity(np.arange(1, 1000), "keV")
    assert len(mat.transmission(energy)) == len(energy)
    assert len(mat.absorption(energy)) == len(energy)
    assert len(mat.mass_attenuation_coefficient(energy)) == len(energy)
    assert len(mat.linear_attenuation_coefficient(energy)) == len(energy)


@pytest.mark.parametrize(
    "material",
    [
        ("Si"),
        ("Au"),
        ("Air (dry)"),
        ("blood"),
        ({"Fe": 0.98, "C": 0.02}),  # steel
        ({"Cu": 0.88, "Sn": 0.12}),  # bronze
        ({"water": 0.97, "Na": 0.015, "Cl": 0.015}),  # salt water
    ],
)
def test_material_scalar_energy(material):
    """If a scalar energy is given then the results should NOT be arrays"""
    mat = Material(material, 1 * u.m)
    energy = 1 * u.keV
    assert isinstance(mat.transmission(energy), float)
    assert isinstance(mat.absorption(energy), float)
    assert mat.mass_attenuation_coefficient(energy).isscalar
    assert mat.linear_attenuation_coefficient(energy).isscalar


@pytest.mark.parametrize(
    "material",
    [
        ("Si"),
        ("Au"),
        ("Air (dry)"),
        ("blood"),
        ({"Fe": 0.98, "C": 0.02}),  # steel
        ({"Cu": 0.88, "Sn": 0.12}),  # bronze
        ({"water": 0.97, "Na": 0.015, "Cl": 0.015}),  # salt water
    ],
)
def test_mass_atten_calculation_in_material(material):
    mat = Material(material, 1 * u.m)
    energy = u.Quantity(np.arange(1, 1000), "keV")
    result1 = mat.mass_attenuation_coefficient(energy)[10]
    result2 = np.average(
        u.Quantity(
            [this_atten.func(energy[10]) for this_atten in mat.mass_attenuation_coefficients]
        ),
        weights=mat.fractional_masses,
    )
    assert np.isclose(result1, result2)


@pytest.mark.parametrize(
    "material_dict",
    [
        ({"Fe": 0.98, "C": 0.02}),  # steel
        ({"Cu": 0.88, "Sn": 0.12}),  # bronze
        ({"water": 0.97, "Na": 0.015, "Cl": 0.015}),  # salt water
    ],
)
def test_dict_input(material_dict):
    assert isinstance(Material(material_dict, 5 * u.mm), Material)


def test_density_calculation():
    a = {"Fe": 0.98, "C": 0.02}
    this_mat = Material(a, 5 * u.m)
    calc_density = get_material_density("Fe") * 0.98 + get_material_density("C") * 0.02
    assert np.isclose(this_mat.density, calc_density)
    a = {"Fe": 0.5, "C": 0.5}
    this_mat = Material(a, 5 * u.m)
    calc_density = get_material_density("Fe") * 0.5 + get_material_density("C") * 0.5
    assert np.isclose(this_mat.density, calc_density)
    a = {"Fe": 1, "C": 1}
    this_mat = Material(a, 5 * u.m)
    calc_density = get_material_density("Fe") * 0.5 + get_material_density("C") * 0.5
    assert np.isclose(this_mat.density, calc_density)
